@@ -363,76 +363,5 @@ public:
 
 #endif /* __ppc__ */
 
-inline UInt32 IOMbufMemoryCursor::getAndResetCoalesceCount()
-{
-    UInt32 cnt = coalesceCount; coalesceCount = 0; return cnt;
-}
-
-inline UInt32
-IOMbufNaturalMemoryCursor::getPhysicalSegments(struct mbuf *packet,
-                       struct IOPhysicalSegment *vector,
-                       UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, false);
-}
-
-inline UInt32
-IOMbufNaturalMemoryCursor::getPhysicalSegmentsWithCoalesce(struct mbuf *packet,
-                                   struct IOPhysicalSegment *vector,
-                                   UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, true);
-}
-
-inline UInt32
-IOMbufBigMemoryCursor::getPhysicalSegments(struct mbuf *packet,
-                       struct IOPhysicalSegment *vector,
-                       UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, false);
-}
-
-inline UInt32
-IOMbufBigMemoryCursor::getPhysicalSegmentsWithCoalesce(struct mbuf *packet,
-                                   struct IOPhysicalSegment *vector,
-                                   UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, true);
-}
-
-inline UInt32
-IOMbufLittleMemoryCursor::getPhysicalSegments(struct mbuf *packet,
-                       struct IOPhysicalSegment *vector,
-                       UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, false);
-}
-
-inline UInt32
-IOMbufLittleMemoryCursor::getPhysicalSegmentsWithCoalesce(struct mbuf *packet,
-                                   struct IOPhysicalSegment *vector,
-                                   UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, true);
-}
-
-#ifdef __ppc__
-inline UInt32
-IOMbufDBDMAMemoryCursor::getPhysicalSegments(struct mbuf *packet,
-                                   struct IODBDMADescriptor *vector,
-                                   UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, false);
-}
-
-inline UInt32
-IOMbufDBDMAMemoryCursor::getPhysicalSegmentsWithCoalesce(struct mbuf *packet,
-                                   struct IODBDMADescriptor *vector,
-                                   UInt32 numVectorSegments = 0)
-{
-    return genPhysicalSegments(packet, vector, numVectorSegments, true);
-}
-#endif /* __ppc__ */
-
 #endif /* !_IOKIT_NETWORK_IOMBUFMEMORYCURSOR_H */
 
